from pvsite_datamodel import DatabaseConnection
from pvsite_datamodel.sqlmodels import LocationSQL


class TestDatabaseConnection:
    """Tests for the DatabaseConnection class."""

    def test_connection(self, engine, sites):
        dbcon = DatabaseConnection(engine.url, echo=False)
        with dbcon.get_session() as session:
            session.query(LocationSQL).first()



"""Test typical ORM queries."""

import datetime as dt

from pvsite_datamodel.sqlmodels import ForecastSQL, ForecastValueSQL, LocationSQL


def _add_forecast_value(session, forecast, power: int, ts):
    session.add(
        ForecastValueSQL(
            forecast_uuid=forecast.forecast_uuid,
            forecast_power_kw=power,
            start_utc=ts,
            end_utc=ts + dt.timedelta(minutes=5),
        ),
    )


def test_get_latest_forecasts(db_session, engine, sites, forecast_valid_site):
    """Get the values of the latest forecast for given pv sites.

    The query looks like this:

    SELECT f.site_uuid, fv.forecast_power_kw, fv.start_utc
    FROM (
        SELECT
        DISTINCT ON (site_uuid)
            forecast_uuid, site_uuid
        FROM forecasts
        where site_uuid IN ('04b3d67f-af48-42f7-ac0f-9d07d4dd84f6', ...)
        ORDER BY site_uuid, timestamp_utc DESC
    ) AS f
    JOIN forecast_values AS fv
      ON fv.forecast_uuid = f.forecast_uuid
    """
    # Get some sites.
    site_uuids = [
        site.location_uuid for site in db_session.query(LocationSQL.location_uuid).limit(2)
    ]

    s1, s2 = site_uuids

    forecast_version = "123"

    # Make sure we have some forecasts in the DB
    s1_f1 = ForecastSQL(
        location_uuid=s1,
        forecast_version=forecast_version,
        timestamp_utc=dt.datetime(2000, 1, 1),
    )
    s1_f2 = ForecastSQL(
        location_uuid=s1,
        forecast_version=forecast_version,
        timestamp_utc=dt.datetime(2000, 1, 2),
    )
    s2_f1 = ForecastSQL(
        location_uuid=s2,
        forecast_version=forecast_version,
        timestamp_utc=dt.datetime(2000, 1, 1),
    )

    db_session.add_all([s1_f1, s1_f2, s2_f1])
    db_session.commit()

    # site 1 forecast 1
    _add_forecast_value(db_session, s1_f1, 1.0, dt.datetime(2000, 1, 1))
    _add_forecast_value(db_session, s1_f1, 2.0, dt.datetime(2000, 1, 1, 1))
    _add_forecast_value(db_session, s1_f1, 3.0, dt.datetime(2000, 1, 1, 2))
    # site 1 forecast 2
    _add_forecast_value(db_session, s1_f2, 4.0, dt.datetime(2000, 1, 1, 2))
    _add_forecast_value(db_session, s1_f2, 5.0, dt.datetime(2000, 1, 1, 3))
    _add_forecast_value(db_session, s1_f2, 6.0, dt.datetime(2000, 1, 1, 4))
    # Site 2 forecast 1
    _add_forecast_value(db_session, s2_f1, 7.0, dt.datetime(2000, 1, 1))
    _add_forecast_value(db_session, s2_f1, 8.0, dt.datetime(2000, 1, 1, 1))
    _add_forecast_value(db_session, s2_f1, 9.0, dt.datetime(2000, 1, 1, 2))
    db_session.commit()

    # Subquery for the forecasts.
    subquery = (
        db_session.query(ForecastSQL)
        .distinct(ForecastSQL.location_uuid)
        .filter(ForecastSQL.location_uuid.in_(site_uuids))
        .order_by(
            ForecastSQL.location_uuid,
            ForecastSQL.timestamp_utc.desc(),
        )
    )

    assert subquery.count() == 2

    # Final query for the forecast values.
    subquery = subquery.subquery()
    values = (
        db_session.query(subquery, ForecastValueSQL)
        .join(ForecastValueSQL)
        .order_by(subquery.c.timestamp_utc, ForecastValueSQL.start_utc)
    )

    expected = [7, 8, 9, 4, 5, 6]
    assert [v.ForecastValueSQL.forecast_power_kw for v in values] == expected



"""Functions for reading user data from the database."""

import logging
from datetime import datetime

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Session, contains_eager

from pvsite_datamodel.sqlmodels import APIRequestSQL, LocationGroupSQL, UserSQL

logger = logging.getLogger(__name__)


def get_user_by_email(session: Session, email: str, make_new_user_if_none: bool = True) -> UserSQL:
    """Get user by email. If user does not exist, make one.

    :param session: database session
    :param email: email of user
    :param make_new_user_if_none: make user with email if doesn't exist
    :return: user object
    """
    user = session.query(UserSQL).filter(UserSQL.email == email).first()

    if user is None and make_new_user_if_none is True:
        logger.info(f"User with email {email} not found, so making one")

        # checking for site_group
        site_group_name = f"site_group_for_{email}"
        site_group = get_site_group_by_name(session=session, site_group_name=site_group_name)

        # make a new user
        stmt = postgresql.insert(UserSQL.__table__)
        stmt = stmt.on_conflict_do_nothing()
        session.execute(
            stmt,
            [{"location_group_uuid": site_group.location_group_uuid, "email": email}],
        )

        # get a new user
        user = session.query(UserSQL).filter(UserSQL.email == email).first()

    return user


# get all users
def get_all_users(session: Session) -> list[UserSQL]:
    """Get all users from the database.

    :param session: database session
    """
    query = session.query(UserSQL)

    query = query.order_by(UserSQL.email.asc())

    users = query.all()

    return users


def get_site_group_by_name(
    session: Session,
    site_group_name: str,
    create_if_none: bool = True,
) -> LocationGroupSQL:
    """Get site group by name. If site group does not exist, make one.

    :param session: database session
    :param site_group_name: name of site group
    :return: site group object
    """
    site_group = (
        session.query(LocationGroupSQL)
        .filter(LocationGroupSQL.location_group_name == site_group_name)
        .first()
    )

    if (site_group is None) and (create_if_none is True):
        logger.info(f"Site group with name {site_group_name} not found, so making one")

        stmt = postgresql.insert(LocationGroupSQL.__table__)
        stmt = stmt.on_conflict_do_nothing()
        session.execute(stmt, [{"location_group_name": site_group_name}])

        site_group = (
            session.query(LocationGroupSQL)
            .filter(LocationGroupSQL.location_group_name == site_group_name)
            .first()
        )

    return site_group


# get all site groups
def get_all_site_groups(session: Session) -> list[LocationGroupSQL]:
    """Get all site groups from the database.

    :param session: database session
    """
    query = session.query(LocationGroupSQL)

    query = query.order_by(LocationGroupSQL.location_group_name.asc())

    site_groups = query.all()

    return site_groups


def get_all_last_api_request(
    session: Session,
    include_in_url: str | None = None,
    exclude_in_url: str | None = None,
    start_datetime: datetime | None = None,
    end_datetime: datetime | None = None,
) -> list[APIRequestSQL]:
    """Get all last api requests for all users.

    :param session: database session
    :param include_in_url: Optional filter to include only URLs containing this string
    :param exclude_in_url: Optional filter to exclude URLs containing this string
    :param start_datetime: only get api requests after start datetime
    :param end_datetime: only get api requests before end datetime
    :return: List of last API requests
    """
    query = (
        session.query(APIRequestSQL)
        .distinct(APIRequestSQL.user_uuid)
        .join(UserSQL)
        .options(contains_eager(APIRequestSQL.user))
        .populate_existing()
        .order_by(APIRequestSQL.user_uuid, APIRequestSQL.created_utc.desc())
    )
    # Apply URL filtering if specified
    if include_in_url is not None:
        query = query.filter(APIRequestSQL.url.like(f"%{include_in_url}%"))

    if exclude_in_url is not None:
        query = query.filter(~APIRequestSQL.url.like(f"%{exclude_in_url}%"))

    if start_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc >= start_datetime)

    if end_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc <= end_datetime)

    return query.all()


def get_api_requests_for_one_user(
    session: Session,
    email: str,
    start_datetime: datetime | None = None,
    end_datetime: datetime | None = None,
    include_in_url: str | None = None,
    exclude_in_url: str | None = None,
) -> list[APIRequestSQL]:
    """Get all api requests for one user.

    :param session: database session
    :param email: user email
    :param start_datetime: only get api requests after start datetime
    :param end_datetime: only get api requests before end datetime
    """
    query = session.query(APIRequestSQL)
    query = query.join(UserSQL).filter(UserSQL.email == email)

    if include_in_url is not None:
        query = query.filter(APIRequestSQL.url.like(f"%{include_in_url}%"))

    if exclude_in_url is not None:
        query = query.filter(~APIRequestSQL.url.like(f"%{exclude_in_url}%"))

    if start_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc >= start_datetime)

    if end_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc <= end_datetime)

    api_requests: [APIRequestSQL] = query.order_by(APIRequestSQL.created_utc.desc()).all()

    return api_requests



import os
import subprocess

from testcontainers.postgres import PostgresContainer


def test_migrations():
    """Run the alembic migrations on a clean database."""
    with PostgresContainer("postgres:14.5") as postgres:
        url = postgres.get_connection_url()

        env = os.environ.copy()
        env["DB_URL"] = url
        result = subprocess.run(["alembic", "upgrade", "head"], env=env) # noqa: S607
        assert result.returncode == 0



"""Test Delete Functions"""

from pvsite_datamodel.sqlmodels import LocationHistorySQL, LocationSQL, UserSQL
from pvsite_datamodel.write.user_and_site import (
    create_site_group,
    create_user,
    delete_site,
    delete_site_group,
    delete_user,
    make_fake_site,
)


# test to delete a site
def test_delete_site(db_session):
    """Test to delete a site."""

    site_1 = make_fake_site(db_session=db_session, ml_id=1)

    # history table should contain a single entry
    hist_size = (
        db_session.query(LocationHistorySQL)
        .filter(LocationHistorySQL.operation_type == "INSERT")
        .count()
    )
    assert hist_size == 1

    # todo add site to site group

    site_group = create_site_group(db_session=db_session, site_group_name="test_site_group")
    site_group.locations.append(site_1)

    site_uuid = site_1.location_uuid

    message = delete_site(session=db_session, site_uuid=site_uuid)

    site = db_session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).first()

    assert site is None
    assert message == f"Location with location uuid {site_uuid} deleted successfully"

    deleted_site = (
        db_session.query(LocationHistorySQL)
        .filter(LocationHistorySQL.operation_type == "DELETE")
        .first()
    )
    assert deleted_site.location_uuid == site_uuid


def test_delete_user(db_session):
    """Test to delete a user."""

    site_group_1 = create_site_group(db_session=db_session)

    user_1 = create_user(
        session=db_session,
        email="test_user@ocf.org",
        site_group_name=site_group_1.location_group_name,
    )

    message = delete_user(session=db_session, email=user_1.email)

    user = db_session.query(UserSQL).filter(UserSQL.email == user_1.email).first()

    assert (
        message == f"User with email {user_1.email} and "
        f"location_group_uuid {user_1.location_group_uuid} deleted successfully"
    )
    assert user is None


def test_delete_site_group(db_session):
    site_group = create_site_group(db_session=db_session, site_group_name="test_site_group")

    message = delete_site_group(session=db_session, site_group_name=site_group.location_group_name)

    assert (
        message == f"Location group with name {site_group.location_group_name} and "
        f"location group uuid {site_group.location_group_uuid} deleted successfully."
    )


def test_delete_site_group_with_users(db_session):
    """Test to delete a site group with users."""

    site_group_1 = create_site_group(db_session=db_session)

    _ = create_user(
        session=db_session,
        email="test_user@gmail.com",
        site_group_name=site_group_1.location_group_name,
    )

    message = delete_site_group(
        session=db_session,
        site_group_name=site_group_1.location_group_name,
    )

    assert (
        message == f"Location group with name {site_group_1.location_group_name} and "
        f"location group uuid {site_group_1.location_group_uuid} has users and cannot be deleted."
    )



from pvsite_datamodel.write.user_and_site import create_site_group, create_user, make_fake_site


def test_one_site_one_user(db_session):
    """
    1. Test that one site with one user works
    """

    site_group = create_site_group(db_session=db_session)
    one_site = make_fake_site(db_session=db_session)
    user = create_user(
        session=db_session,
        email="test_user@gmail.com",
        site_group_name=site_group.location_group_name,
    )

    # add site to site group
    site_group.locations.append(one_site)

    assert len(user.location_group.locations) == 1


def test_two_site_one_user(db_session):
    """
    2. Test that one site with one user works
    """

    site_group = create_site_group(db_session=db_session)
    site_1 = make_fake_site(db_session=db_session, ml_id=1)
    site_2 = make_fake_site(db_session=db_session, ml_id=2)
    user = create_user(
        session=db_session,
        email="test_user@gmail.com",
        site_group_name=site_group.location_group_name,
    )

    # add site to site group
    site_group.locations.append(site_1)
    site_group.locations.append(site_2)

    assert len(user.location_group.locations) == 2


def test_one_site_two_user(db_session):
    """
    3. Test that one site with two users works
    """

    site_group = create_site_group(db_session=db_session)
    site_1 = make_fake_site(db_session=db_session, ml_id=1)
    user_1 = create_user(
        session=db_session,
        email="test_user_1@gmail.com",
        site_group_name=site_group.location_group_name,
    )
    user_2 = create_user(
        session=db_session,
        email="test_user_2@gmail.com",
        site_group_name=site_group.location_group_name,
    )

    # add site to site group
    site_group.locations.append(site_1)

    assert len(user_1.location_group.locations) == 1
    assert len(user_2.location_group.locations) == 1


def test_two_site_two_user(db_session):
    """
    4. Test that has two site with two users
    """

    site_group = create_site_group(db_session=db_session)
    site_1 = make_fake_site(db_session=db_session, ml_id=1)
    site_2 = make_fake_site(db_session=db_session, ml_id=2)
    user_1 = create_user(
        session=db_session,
        email="test_user_1@gmail.com",
        site_group_name=site_group.location_group_name,
    )
    user_2 = create_user(
        session=db_session,
        email="test_user_2@gmail.com",
        site_group_name=site_group.location_group_name,
    )

    # add site to site group
    site_group.locations.append(site_1)
    site_group.locations.append(site_2)

    assert len(user_1.location_group.locations) == 2
    assert len(user_2.location_group.locations) == 2


def test_three_site_two_user_and_ocf_see_everything(db_session):
    """
    5. Test that three site with two user works
    """

    site_group = create_site_group(db_session=db_session)
    site_group_ocf = create_site_group(db_session=db_session, site_group_name="OCF")
    site_1 = make_fake_site(db_session=db_session, ml_id=1)
    site_2 = make_fake_site(db_session=db_session, ml_id=2)
    site_3 = make_fake_site(db_session=db_session, ml_id=3)
    user_1 = create_user(
        session=db_session,
        email="test_user_1@gmail.com",
        site_group_name=site_group.location_group_name,
    )
    user_2 = create_user(
        session=db_session,
        email="test_user_2@gmail.com",
        site_group_name=site_group.location_group_name,
    )
    user_ocf = create_user(
        session=db_session,
        email="test_user_ocf@gmail.com",
        site_group_name=site_group_ocf.location_group_name,
    )

    # add site to site group
    site_group.locations.append(site_1)
    site_group.locations.append(site_2)
    site_group_ocf.locations.append(site_1)
    site_group_ocf.locations.append(site_2)
    site_group_ocf.locations.append(site_3)

    assert len(user_1.location_group.locations) == 2
    assert len(user_2.location_group.locations) == 2
    assert len(user_ocf.location_group.locations) == 3
